/*
 * Title        :  ResourceMonitor.java
 * Package      :  org.gridbus.gridscape.resourcemonitor
 * Project      :  Gridscape Resource Monitor
 * Description	:  		
 * Created on   :  Feb 9, 2006
 * Author		:  Hussein Gibbins (hag@cs.mu.oz.au)
 * Copyright    :  (c) 2006, Grid Computing and Distributed Systems Laboratory, 
 * 					   Dept. of Computer Science and Software Engineering,
 * 					   University of Melbourne, Australia.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later  version.
 * See the GNU General Public License (http://www.gnu.org/copyleft/gpl.html)for more details.
 * 
 */

package org.gridbus.gridscape.resourcemonitor;

import java.lang.reflect.Constructor;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;
import org.gridbus.gridscape.core.Resource;
import org.gridbus.gridscape.core.ResourceInfo;
import org.gridbus.gridscape.core.data.DBQuerier;

/**
 * The resource monitor application is used to update the status of resources
 * in the Gridscape database.  The monitor performs the following:
 * 
 * 1. reads all resources from the database
 * 2. for each resource start a new thread and perform in the thread:
 *   2.1. create an info gatherer based on the resource type
 *   2.2. use the info gatherer to query the resource info
 *   2.3. update the resource info in the database
 * 3. sleep for a while and then repeat from step 1.
 * 
 * @author hag
 *
 */
public class ResourceMonitor {

    private static Hashtable threads = new Hashtable();
    
    private static long POLL_TIME = 200000;
    private static final long MIN_POLL_TIME = 50000;
    private static final long THREAD_GRACE_PERIOD = 2000;
//    private static final long THREAD_TIMEOUT = 30000;
    
    /**
     * @param args
     */
    public static void main(String[] args) {

        if (parseCLI(args) == false) {
            System.exit(1);
        }

        System.out.println("Starting Resource Monitor... ");
        
        Map resources = null;
        DBQuerier querier = null;
        
        try {
            querier = new DBQuerier();
        } catch (Exception e) {
            System.err.println("Cannot connect to database. Exiting.");
            System.exit(1);
        }

        System.out.println("Resource Monitor Started... ");

        while (true) {
            try {
                // read resources from db
                resources = querier.getResources();

                for (Iterator it = resources.values().iterator();it.hasNext();) {
                    Resource r = (Resource)it.next();
                    Thread tmp = (Thread)threads.get(r.getHostname());
                    if (tmp != null && tmp.isAlive()) {
                        System.out.println("**** still gathering info from "+r.getHostname()+". Aborting and performing new query.");
                        try {
                            tmp.interrupt();
                        } catch (Exception e) { }
                    }
                    System.out.println("**** new query on "+r.getHostname()+".");
                    Thread t = new Thread(new InfoGathererThread(r),r.getHostname());
                    t.start();
                    threads.put(r.getHostname(),t);
                    Thread.sleep(THREAD_GRACE_PERIOD);
                }
                Thread.sleep(POLL_TIME);

            } catch (Exception e) {
                System.err.println("Error updating resource information");
                e.printStackTrace();
            }
        }
    }

    private static boolean parseCLI(String[] args) {
        
//      create Options object
        Options options = new Options();
        

        options.addOption(OptionBuilder.withLongOpt("help")
                                        .withDescription("prints this message")
                                        .create("h"));

        options.addOption( OptionBuilder.withLongOpt("poll-time")
                                        .withArgName("time")
                                        .hasArg()
                                        .withValueSeparator('=')
                                        .withDescription("time to wait (in milliseconds) between queries to the same resource.  Default 200000")
                                        .create("p"));

        // create the parser
        CommandLineParser parser = new PosixParser();
        HelpFormatter formatter = new HelpFormatter();

        try {
            // parse the command line arguments
            CommandLine line = parser.parse( options, args );

            if( line.hasOption("h") ) {
                formatter.printHelp("java ResourceMonitor", options);
                return false;
            }
            
            if ( line.hasOption("p")) {
                POLL_TIME = Long.parseLong(line.getOptionValue("p"));
                if (POLL_TIME < MIN_POLL_TIME) POLL_TIME = MIN_POLL_TIME;
            }
        } catch( ParseException e ) {
            System.err.println(e.getMessage()+"\n");
            formatter.printHelp("java ResourceMonitor", options);
            return false;
        }
        
        return true;
    }
    
///////////////////////////////////////////////////
//// INFO GATHERER STUFF
///////////////////////////////////////////////////
    
    private static InfoGatherer getInfoGatherer(Resource r) {
        try {
            //get the appropriate class based on the element name.
            Class clas = Class.forName("org.gridbus.gridscape.resourcemonitor.plugins."+r.getType()+"InfoGatherer");
            
            //get the default constructor for that class. 
            Constructor cons = clas.getConstructor((Class[])null);
            //create an object of that class.
            Object newObj = cons.newInstance((Object[])null);
            
            if (newObj instanceof InfoGatherer) {
                return (InfoGatherer)newObj;
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return null;
    }

    final static class InfoGathererThread implements Runnable {
    
        private Resource r = null;

        /**
         * @param r
         */
        public InfoGathererThread(Resource r) {
            this.r = r;
        }

        public void run() {
            DBQuerier querier = null;
            try {
                // discover properties for each result
                InfoGatherer ig = getInfoGatherer(r);
                // if not null query in a thread
                if (ig != null) {
                    ResourceInfo info = ig.gatherInfo(r);
                    querier = new DBQuerier();
                    if (info != null) {
                        querier.setOnline(r.getID(),true);
                        // update info into db 
                        querier.updateResourceInfo(r.getID(),info);
                    } else {
                        // set online to false
                        querier.setOnline(r.getID(),false);
                    }
                } else {
                    System.err.println("Bad gatherer type!");
                }
            } catch (Exception e) {
                System.err.println("Unable to update database.");
            }
        }
    }
}
